"""
Module consisting of explorers and tracks that connect arbitrary paraview
pipelines to cinema stores.
"""
from __future__ import absolute_import

import numpy as np
import math
from .. import explorers as explorers
from ...images.camera_utils import convert_pose_to_camera

import paraview.simple as simple
from paraview import numpy_support as numpy_support


class ValueMode():
    INVERTIBLE_LUT = 1
    FLOATING_POINT = 2


def printView(view):
    """
    For debugging/demonstrating frustums
    spits out the coordinates of the four camera near plane corners
    load them up in paraview
    """
    r = view.GetRenderer()
    r.SetViewPoint(-1, -1, 0)
    r.ViewToWorld()
    ll = r.GetWorldPoint()
    r.SetViewPoint(1, -1, 0)
    r.ViewToWorld()
    lr = r.GetWorldPoint()
    r.SetViewPoint(1, 1, 0)
    r.ViewToWorld()
    ur = r.GetWorldPoint()
    r.SetViewPoint(-1, 1, 0)
    r.ViewToWorld()
    ul = r.GetWorldPoint()
    print(ll[0], ll[1], ll[2])
    print(lr[0], lr[1], lr[2])
    print(ur[0], ur[1], ur[2])
    print(ul[0], ul[1], ul[2])
    # print (math.sqrt((ur[0]-ll[0])*(ur[0]-ll[0])+
    #                  (ur[1]-ll[1])*(ur[1]-ll[1])+
    #                  (ur[2]-ll[2])*(ur[2]-ll[2])))


class ImageExplorer(explorers.Explorer):
    """
    An Explorer that connects a ParaView script's views to a store.
    Basically it iterates over the parameters and for each unique combination
    it tells ParaView to make an image and saved the result into the store.
    """

    def __init__(self,
                 store, parameters, tracks,
                 view=None,
                 iSave=True):
        super(ImageExplorer, self).__init__(store, parameters, tracks)
        self.view = view
        self.CaptureDepth = False
        self.CaptureLuminance = False
        self.CaptureValues = False
        self.iSave = iSave
        self.UsingGL2 = False
        # Using float rasters for value arrays by default. vtkPVRenderView
        # will fall back to INVERTIBLE_LUT if the required extensions are
        # not supported.
        self.ValueMode = ValueMode().FLOATING_POINT
        self.CheckFloatSupport = True

        if self.view:
            try:
                # TODO We keep LuminancePass purposely off. There is a bug when
                # in which it seems to remain attached even after StopCapture.
                # backend = self.view.GetRenderWindow().GetRenderingBackend()
                # self.UsingGL2 = True if (backend == "OpenGL2") else False
                pass
            except AttributeError:
                pass
        if self.UsingGL2:
            def rgb2grey(rgb, height, width):
                # pass through Diffuse lum term
                as_grey = np.dot(rgb[..., :3], [0.0, 1.0, 0.0])
                res = as_grey.reshape(height, width).astype('uint8')
                return res
            self.rgb2grey = rgb2grey
        else:
            def rgb2grey(rgb, height, width):
                as_grey = np.dot(rgb[..., :3], [0.299, 0.587, 0.114])
                res = as_grey.reshape(height, width).astype('uint8')
                return res
            self.rgb2grey = rgb2grey

    def enableFloatValues(self, enable):
        self.CheckFloatSupport = True if enable else False
        if enable:
            self.ValueMode = ValueMode().FLOATING_POINT
        else:
            self.ValueMode = ValueMode().INVERTIBLE_LUT

    def insert(self, document):
        """overridden to use paraview to generate an image and create a
        the document for it"""
        if not self.view:
            return

        if self.CaptureDepth:
            # Depth capture
            rep = simple.GetRepresentation()
            wasVol = rep.Representation == 'Volume'
            if wasVol:
                rep.Representation = 'Surface'
            simple.Render()
            image = self.view.CaptureDepthBuffer()
            idata = numpy_support.vtk_to_numpy(image) * 256
            rw = self.view.GetRenderWindow()
            width, height = rw.GetSize()
            try:
                imageslice = idata.reshape(height, width)
            except ValueError:
                imageslice = None

            if wasVol:
                rep.Representation = 'Volume'

            document.data = imageslice
            self.CaptureDepth = False
            # printView(self.view)

        else:
            imageslice = None

            if self.CaptureLuminance:
                if self.UsingGL2:
                    # TODO: needs a fix for luminance pass
                    imageslice = self.captureWindowRGB()
                else:  # GL1 support
                    try:
                        rep = simple.GetRepresentation()
                        if rep is not None:
                            rep.DiffuseColor = [1, 1, 1]
                            rep.ColorArrayName = None
                    except ValueError:
                        pass
                    idata, width, height = self.captureWindowAsNumpy()
                    idata = self.rgb2grey(idata, height, width)
                    imageslice = np.dstack((idata, idata, idata))

                document.data = imageslice

            elif self.CaptureValues:  # Value capture
                if self.ValueMode is ValueMode().INVERTIBLE_LUT:
                    imageslice = self.captureWindowRGB()

                elif self.ValueMode is ValueMode().FLOATING_POINT:
                    simple.Render()
                    image = self.view.GetValuesFloat()
                    idata = numpy_support.vtk_to_numpy(image)
                    idataMin = idata.min() if len(idata) > 0 else 0
                    idataMax = idata.max() if len(idata) > 0 else 0
                    self.updateRange(self.view.ArrayNameToDraw,
                                     self.view.ArrayComponentToDraw,
                                     [idataMin, idataMax])
                    rw = self.view.GetRenderWindow()
                    width, height = rw.GetSize()
                    try:
                        imageslice = idata.reshape(height, width)
                    except ValueError:
                        imageslice = None

                document.data = imageslice

            else:  # Capture color image
                imageslice = self.captureWindowRGB()
                document.data = imageslice

        if self.iSave:
            super(ImageExplorer, self).insert(document)

    def captureWindowRGB(self):
        idata, width, height = self.captureWindowAsNumpy()
        imageslice = idata.reshape(height, width, 3)
        return imageslice

    def captureWindowAsNumpy(self):
        image = self.view.CaptureWindow(1)
        ext = image.GetExtent()
        width = ext[1] - ext[0] + 1
        height = ext[3] - ext[2] + 1
        imagescalars = image.GetPointData().GetScalars()
        idata = numpy_support.vtk_to_numpy(imagescalars)
        image.UnRegister(None)
        return idata, width, height

    def setDrawMode(self, choice, **kwargs):
        """ helper for Color tracks so that they can cause ParaView to
        render in the right mode."""
        if choice == 'color':
            self.view.StopCaptureValues()
            if self.UsingGL2:
                self.view.StopCaptureLuminance()
            self.CaptureDepth = False
            self.CaptureLuminance = False
            self.CaptureValues = False
        if choice == 'luminance':
            self.view.StopCaptureValues()
            if self.UsingGL2:
                self.view.StartCaptureLuminance()
            self.CaptureDepth = False
            self.CaptureLuminance = True
            self.CaptureValues = False
        if choice == 'depth':
            self.view.StopCaptureValues()
            if self.UsingGL2:
                self.view.StopCaptureLuminance()
            self.CaptureDepth = True
            self.CaptureLuminance = False
            self.CaptureValues = False
        if choice == 'value':
            if self.UsingGL2:
                self.view.StopCaptureLuminance()

            if self.CaptureValues is not True:
                # this is a workaround for a bug in VTK
                # in valuepass, we need the arrays we will render,
                # but with CompositePolyDataMapper we need to render
                # to find out what arrays are needed
                self.view.StopCaptureValues()
                simple.Render()

            self.view.DrawCells = kwargs['field']
            self.view.ArrayNameToDraw = kwargs['name']
            self.view.ArrayComponentToDraw = kwargs['component']
            self.view.ScalarRange = kwargs['range']

            self.view.StartCaptureValues()
            self.view.SetValueRenderingMode(self.ValueMode)
            # Ensure context support
            if (self.CheckFloatSupport and
                    self.ValueMode == ValueMode().FLOATING_POINT):
                # Force a proxy update to make sure the rendering mode is set
                self.view.UpdateVTKObjects()
                simple.Render()
                self.ValueMode = self.view.GetValueRenderingMode()
                self.CheckFloatSupport = False
                self.store.add_metadata(
                    {'value_mode': self.ValueMode})

            self.CaptureDepth = False
            self.CaptureLuminance = False
            self.CaptureValues = True

    def finish(self):
        super(ImageExplorer, self).finish()
        # TODO: actually record state in init and restore here, for now just
        # make an assumption
        self.view.StopCaptureValues()
        if self.UsingGL2:
            self.view.StopCaptureLuminance()
        try:
            simple.Show()
            simple.Render()
        except RuntimeError:
            pass

    def updateRange(self, name, component, drange):
        fname = name + "_" + str(component)
        plist = self.store.parameter_list
        for pname, param in plist.items():
            if 'valueRanges' in param:
                # now we know it is a color type parameter
                vrange = param['valueRanges'][fname]
                lrange = list(vrange)
                updated = False
                if drange[0] < vrange[0]:
                    updated = True
                    lrange[0] = drange[0]
                if drange[1] > vrange[1]:
                    updated = True
                    lrange[1] = drange[1]
                if updated:
                    param['valueRanges'][fname][0] = float(lrange[0])
                    param['valueRanges'][fname][1] = float(lrange[1])


class Camera(explorers.Track):
    """
    A track that connects a ParaView script's camera to the phi and theta
    tracks. This allows the creation of spherical camera stores where the
    user can view the data from many points around it.
    """
    def __init__(self, center, axis, distance, view):
        super(Camera, self).__init__()
        try:
            # Z => 0 | Y => 2 | X => 1
            self.offset = (axis.index(1) + 1) % 3
        except ValueError:
            raise Exception("Rotation axis not supported", axis)
        self.center = center
        self.distance = distance
        self.view = view

    def execute(self, document):
        """moves camera into position for the current phi, theta value"""
        theta = document.descriptor['theta']
        phi = document.descriptor['phi']

        theta_rad = float(theta) / 180.0 * math.pi
        phi_rad = float(phi) / 180.0 * math.pi
        pos = [
            math.cos(phi_rad) * self.distance * math.cos(theta_rad),
            math.sin(phi_rad) * self.distance * math.cos(theta_rad),
            math.sin(theta_rad) * self.distance
            ]
        up = [
            + math.cos(phi_rad) * math.sin(theta_rad),
            - math.sin(phi_rad) * math.sin(theta_rad),
            + math.cos(theta_rad)
            ]
        for i in range(self.offset):
            pos.insert(0, pos.pop())
            up.insert(0, up.pop())
        pos[0] = float(self.center[0]) - pos[0]
        pos[1] = float(self.center[1]) + pos[1]
        pos[2] = float(self.center[2]) + pos[2]

        self.view.CameraPosition = pos
        self.view.CameraViewUp = up
        self.view.CameraFocalPoint = self.center


class PoseCamera(explorers.Track):
    """
    A track that connects a ParaView script's pose track which has 3x3
    camera orientations in it. Also takes in camera_eye, _at and _up
    arrays, which designate the initial position at each timestep, from
    which the camera moves according to each pose.
    """
    def __init__(self, view, camType, store):
        super(PoseCamera, self).__init__()
        self.view = view
        self.camType = camType
        self.store = store

    def execute(self, document):
        """moves camera into position for the current phi, theta value"""
        pose = document.descriptor['pose']

        # at every timestep paraview records camera location
        # use that as point to spin around
        iPosition = self.store.metadata['camera_eye'][-1]
        iFocalPoint = self.store.metadata['camera_at'][-1]
        iViewUp = self.store.metadata['camera_up'][-1]

        newp, newf, newv = convert_pose_to_camera(
            iPosition, iFocalPoint, iViewUp, pose, self.camType)
        self.view.GetActiveCamera().SetPosition(newp)
        self.view.GetActiveCamera().SetFocalPoint(newf)
        self.view.GetActiveCamera().SetViewUp(newv)


class Slice(explorers.Track):
    """
    A track that connects a slice filter to a scalar valued parameter.
    """
    def __init__(self, parameter, filt):
        super(Slice, self).__init__()

        self.parameter = parameter
        self.slice = filt

    def prepare(self, explorer):
        super(Slice, self).prepare(explorer)

    def execute(self, doc):
        if self.parameter in doc.descriptor:
            o = doc.descriptor[self.parameter]
            self.slice.SliceOffsetValues = [o]


class Contour(explorers.Track):
    """
    A track that connects a contour filter to a scalar valued parameter.
    """
    def __init__(self, parameter, filt):
        super(Contour, self).__init__()
        self.parameter = parameter
        self.contour = filt
        self.control = 'Isosurfaces'

    def prepare(self, explorer):
        super(Contour, self).prepare(explorer)

    def execute(self, doc):
        if self.parameter in doc.descriptor:
            o = doc.descriptor[self.parameter]
            self.contour.SetPropertyWithName(self.control, [o])


class Clip(explorers.Track):
    """
    A track that connects a clip filter to a scalar valued parameter.
    """
    def __init__(self, argument, clip):
        super(Clip, self).__init__()
        self.argument = argument
        self.clip = clip

    def prepare(self, explorer):
        super(Clip, self).prepare(explorer)

    def execute(self, doc):
        if self.argument in doc.descriptor:
            o = doc.descriptor[self.argument]
            self.clip.UseValueAsOffset = True
            self.clip.Value = o


class Templated(explorers.Track):
    """
    A track that connects any type of filter to a scalar valued
    parameter. To use pass in a source proxy (aka filter)
    and the name of method (aka property) to be called on it.
    """
    def __init__(self, parameter, filt, methodName):
        explorers.Track.__init__(self)

        self.parameter = parameter
        self.filt = filt
        self.methodName = methodName

    def execute(self, doc):
        o = doc.descriptor[self.parameter]
        self.filt.SetPropertyWithName(self.methodName, [o])


class ColorList():
    """
    A helper that creates a dictionary of color controls for ParaView.
    The Color track takes in a color name from the Explorer and looks
    up into a ColorList to determine exactly what needs to be set to
    apply the color.
    """
    def __init__(self):
        self._dict = {}

    def AddSolidColor(self, name, RGB):
        self._dict[name] = {'type': 'rgb', 'content': RGB}

    def AddLUT(self, field, name, comp):
        self._dict[name] = {'type': 'lut', 'field': field,
                            'name': name, 'comp': comp}

    def AddDepth(self, name):
        self._dict[name] = {'type': 'depth'}

    def AddLuminance(self, name):
        self._dict[name] = {'type': 'luminance'}

    def AddValueRender(self, name, field, arrayname, component, range):
        self._dict[name] = {'type': 'value',
                            'field': field,
                            'arrayname': arrayname,
                            'component': component,
                            'range': range}

    def getColor(self, name):
        if name in self._dict:
            return self._dict[name]
        else:
            return {'type': 'none'}


class Color(explorers.Track):
    """
    A track that connects a parameter to color controls.
    """
    def __init__(self, parameter, colorlist, rep):
        super(Color, self).__init__()
        self.parameter = parameter
        self.colorlist = colorlist
        self.rep = rep
        self.imageExplorer = None

    def execute(self, doc):
        """tells ParaView to color the object we've been assigned
        using the color definition we've been given that corresponds
        to the value we've been assigned to watch in the doc.descriptor"""
        if self.parameter not in doc.descriptor:
            return
        if self.rep is None:
            return
        o = doc.descriptor[self.parameter]
        spec = self.colorlist.getColor(o)
        if spec['type'] == 'rgb':
            self.rep.DiffuseColor = spec['content']
            self.rep.ColorArrayName = None
            if self.imageExplorer:
                self.imageExplorer.setDrawMode('color')
        if spec['type'] == 'lut':
            aname = spec['name']
            aname = aname[0:aname.rfind("_")]
            simple.ColorBy(self.rep, (spec['field'], aname, spec['comp']))
            if self.imageExplorer:
                self.imageExplorer.setDrawMode('color')
        if spec['type'] == 'depth':
            if self.imageExplorer:
                self.imageExplorer.setDrawMode('depth')
        if spec['type'] == 'luminance':
            if self.imageExplorer:
                self.imageExplorer.setDrawMode('luminance')
        if spec['type'] == 'value':
            if self.imageExplorer:
                self.imageExplorer.setDrawMode("value",
                                               field=spec['field'],
                                               name=spec['arrayname'],
                                               component=spec['component'],
                                               range=spec['range'])


class SourceProxyInLayer(explorers.LayerControl):
    """
    A track that turns on and off an source proxy in a layer
    """
    def showme(self):
        simple.SetActiveSource(self.proxy)
        self.representation.Visibility = 1

    def hideme(self):
        self.representation.Visibility = 0

    def __init__(self, parameter, representation, proxy):
        super(SourceProxyInLayer, self).__init__(
            parameter, self.showme, self.hideme)
        self.representation = representation
        self.proxy = proxy



import datetime as dt
from paraview import servermanager
from paraview.simple import *
from paraview.benchmark import *
#import logbase, logparser

logbase.maximize_logs()
records = []
n0 = dt.datetime.now()

def get_render_view(size):
    '''Similar to GetRenderView except if a new view is created, it's
    created with the specified size instead of having t resize afterwards
    '''
    view = active_objects.view
    if not view:
        # it's possible that there's no active view, but a render view exists.
        # If so, locate that and return it (before trying to create a new one).
        view = servermanager.GetRenderView()
    if not view:
        view = CreateRenderView(ViewSize=size)
    return view


def save_render_buffer(fname):
    '''Similar to SaveScreenshot except a re-render will not be triggered'''
    from vtkmodules.vtkRenderingCore import vtkWindowToImageFilter
    w = GetRenderView().SMProxy.GetRenderWindow()
    w2i = vtkWindowToImageFilter()
    w2i.ReadFrontBufferOff()
    w2i.ShouldRerenderOff()
    w2i.SetInput(w)
    w2i.Modified()
    png = PNGWriter()
    png.Input = w2i.GetOutput()
    png.FileName = fname
    png.UpdatePipeline()


def flush_render_buffer():
    '''When running as a single process use the WindowToImage filter to
    force a framebuffer read.  This bypasses driver optimizations that
    perform lazy rendering and allows you to get actual frame rates for
    a single process with a GPU.  Multi-process doesn't need this since
    compositing forces the frame buffer read.
    '''

    # If we're not using off-screen rendering then we can bypass this since
    # the frame buffer display will force a GL flush
    w = GetRenderView().SMProxy.GetRenderWindow()
    if not w.GetOffScreenRendering():
        return

    from vtkmodules.vtkParallelCore import vtkMultiProcessController
    from vtkmodules.vtkRenderingCore import vtkWindowToImageFilter

    # If we're using MPI we can also bypass this since compositing will
    # for a GL flush
    controller = vtkMultiProcessController.GetGlobalController()
    if controller.GetNumberOfProcesses() > 1:
        return

    # Force a GL flush by retrieving the frame buffer image
    w2i = vtkWindowToImageFilter()
    w2i.ReadFrontBufferOff()
    w2i.ShouldRerenderOff()
    w2i.SetInput(w)
    w2i.Modified()
    w2i.Update()


def memtime_stamp():
    global records
    global n0
    m = logbase.get_memuse()
    n1 = dt.datetime.now()
    et = n1 - n0
    print(et, m)
    n0 = n1
    records.append([et, m])


def run(output_basename='log', dimension=100, view_size=(1920, 1080),
        num_frames=10, save_logs=True, ospray=False):

    from vtkmodules.vtkParallelCore import vtkMultiProcessController
    from vtkmodules.vtkCommonSystem import vtkTimerLog

    controller = vtkMultiProcessController.GetGlobalController()

    view = get_render_view(view_size)
    if ospray:
        view.EnableRayTracing = 1

    print('Generating wavelet')
    wavelet = Wavelet()
    d2 = dimension//2
    wavelet.WholeExtent = [-d2, d2, -d2, d2, -d2, d2]
    wavelet.Maximum = 100.0
    waveletDisplay = Show()
    waveletDisplay.SetRepresentationType('Volume')

    print('Repositioning initial camera')
    c = GetActiveCamera()
    c.Azimuth(22.5)
    c.Elevation(22.5)

    print('Rendering first frame')
    Render()

    print('Saving frame 0 screenshot')
    import math
    fdigits = int(math.ceil(math.log(num_frames, 10)))
    frame_fname_fmt = output_basename + '.scene.f%(f)0' + str(fdigits) + 'd.png'
    SaveScreenshot(frame_fname_fmt % {'f': 0})

    print('Gathering geometry counts')
    vtkTimerLog.MarkStartEvent('GetViewItemStats')
    num_voxels = 0
    for r in view.Representations:
        num_voxels += r.GetRepresentedDataInformation().GetNumberOfCells()
    vtkTimerLog.MarkEndEvent('GetViewItemStats')

    print('Beginning benchmark loop')
    deltaAz = 45.0 / num_frames
    deltaEl = 45.0 / num_frames
    memtime_stamp()
    fpsT0 = dt.datetime.now()
    for frame in range(1, num_frames):
        c.Azimuth(deltaAz)
        c.Elevation(deltaEl)
        Render()
        flush_render_buffer()
        memtime_stamp()
    fpsT1 = dt.datetime.now()

    if controller.GetLocalProcessId() == 0:
        if save_logs:
            # Save the arguments this was executed with
            with open(output_basename + '.args.txt', 'w') as argfile:
                argfile.write(str({
                    'output_basename': output_basename,
                    'dimension': dimension,
                    'view_size': view_size,
                    'num_frames': num_frames,
                    'ospray' : ospray,
                    'save_logs': save_logs}))

            # Save the memory statistics collected
            with open(output_basename + '.mem.txt', 'w') as ofile:
                ofile.write('\n'.join([str(x) for x in records]))

        # Process frame timing statistics
        logparser.summarize_results(num_frames, (fpsT1-fpsT0).total_seconds(),
                                    num_voxels, 'Voxels', save_logs,
                                    output_basename)


def main(argv):
    import argparse
    parser = argparse.ArgumentParser(
        description='Benchmark ParaView geometry rendering')
    parser.add_argument('-o', '--output-basename', default='log', type=str,
                        help='Basename to use for generated output files')
    parser.add_argument('-d', '--dimension', default=100, type=int,
                        help='The dimension of each side of the cubic volume')
    parser.add_argument('-v', '--view-size', default=[400, 400],
                        type=lambda s: [int(x) for x in s.split(',')],
                        help='View size used to render')
    parser.add_argument('-f', '--frames', default=10, type=int,
                        help='Number of frames')
    parser.add_argument('-y', '--ospray', action='store_true',
                        help='Use OSPRAY to render')

    args = parser.parse_args(argv)

    options = servermanager.vtkProcessModule.GetProcessModule().GetOptions()
    url = options.GetServerURL()
    if url:
        import re
        m = re.match('([^:/]*://)?([^:]*)(:([0-9]+))?', url)
        if m.group(4):
            Connect(m.group(2), m.group(4))
        else:
            Connect(m.group(2))

    run(output_basename=args.output_basename, dimension=args.dimension,
        view_size=args.view_size, num_frames=args.frames, ospray=args.ospray)

if __name__ == "__main__":
    import sys
    main(sys.argv[1:])



from __future__ import absolute_import, print_function

from functools import wraps
from xml.dom.minidom import parseString
from xml.parsers.expat import ExpatError
from inspect import getargspec

import sys

def _count(values):
    """internal: returns the `number-of-elements` for the argument.
       if argument is a list or type, it returns its `len`, else returns 0 for None
       and 1 for non-None."""
    if type(values) == list or type(values) == tuple:
        return len(values)
    elif values is None:
        return 0
    else:
        return 1

def _stringify(values):
    """internal method: used to convert values to a string suitable for an xml attribute"""
    if type(values) == list or type(values) == tuple:
        return " ".join([str(x) for x in values])
    elif type(values) == type(True):
        return "1" if values else "0"
    else:
        return str(values)

def _generate_xml(attrs, nested_xmls=[]):
    """internal: used to generate an XML string from the arguments.
    `attrs` is a dict with attributes specified using (key, value) for the dict.
       `type` key in `attrs` is treated as the XML tag name.

    `nested_xmls` is a list of strings that get nested in the resulting xmlstring
        returned by this function.
    """
    d = {}
    d["type"] = attrs.pop("type")

    attr_items = filter(lambda item : item[1] is not None, attrs.items())
    d["attrs"] = "\n".join(["%s=\"%s\"" % (x, _stringify(y)) for x,y in attr_items])
    d["nested_xmls"] = "\n".join(nested_xmls)
    xml = """<{type} {attrs}> {nested_xmls} </{type}>"""
    return xml.format(**d)


def _undecorate(func):
    """internal: Traverses through nested decorated objects to return the original
    object. This is not a general mechanism and only supports decorator chains created
    via `_create_decorator`."""
    if hasattr(func, "_pv_original_func"):
        return _undecorate(func._pv_original_func)
    return func

def _create_decorator(kwargs={}, update_func=None, generate_xml_func=None):
    """internal: used to create decorator for class or function objects.

    `kwargs`: these are typically the keyword arguments passed to the decorator itself.
              must be a `dict`. The decorator will often update the dict to have
              defaults or overrides to the parameters passed to the decorator, as appropriate.

    `update_func`: if non-None, must be a callable that takes 2 arguments `(decoratedobj, kwargs)`.
                   The purpose of this callback is to update the kwargs (and return updated version)
                   for required-yet-missing attributes that can be deduced by introspecting
                   the `decoratedobj`.

    `generate_xml_func`: must be non-None and must be a callable that takes 2 arguments `(decoratedobj, kwargs)`.
                   The kwargs can be expected to have been updated by calling `update_func`, if non-None.
                   The callback typically adds an attribute on the decoratedobj for further processing later.
    """
    attrs = kwargs.copy()
    def decorator(func):
        original_func = _undecorate(func)
        if update_func is not None:
            updated_attrs = update_func(original_func, attrs)
        else:
            updated_attrs = attrs
        generate_xml_func(original_func, updated_attrs)
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        setattr(wrapper, "_pv_original_func", original_func)
        return wrapper
    return decorator


class smproperty(object):
    """
    Provides decorators for class methods that are to be exposed as
    server-manager properties in ParaView. Only methods that are decorated
    using one of the available decorators will be exposed be accessible to ParaView
    UI or client-side Python scripting API.
    """
    @staticmethod
    def _append_xml(func, xml):
        pxmls = []
        if hasattr(func, "_pvsm_property_xmls"):
            pxmls = func._pvsm_property_xmls
        pxmls.append(xml)
        setattr(func, "_pvsm_property_xmls", pxmls)

    @staticmethod
    def _generate_xml(func, attrs):
        nested_xmls = []
        if hasattr(func, "_pvsm_domain_xmls"):
            for d in func._pvsm_domain_xmls:
                nested_xmls.append(d)
            delattr(func, "_pvsm_domain_xmls")
        if hasattr(func, "_pvsm_hints_xmls"):
            hints = []
            for h in func._pvsm_hints_xmls:
                hints.append(h)
            nested_xmls.append(_generate_xml({"type":"Hints"}, hints))
            delattr(func, "_pvsm_hints_xmls")

        pxml = _generate_xml(attrs, nested_xmls)
        smproperty._append_xml(func, pxml)

    @staticmethod
    def _update_property_defaults(func, attrs):
        """Function used to populate default attribute values for missing attributes
           on a for a property"""
        # determine unspecified attributes based on the `func`
        if attrs.get("name", None) is None:
            attrs["name"] = func.__name__
        if attrs.get("command", None) is None:
            attrs["command"] = func.__name__
        return attrs

    @staticmethod
    def _update_vectorproperty_defaults(func, attrs):
        """Function used to populate default attribute values for missing attributes
           on a vector property"""
        attrs = smproperty._update_property_defaults(func, attrs)

        if attrs.get("number_of_elements", None) is None:
            attrs["number_of_elements"] = len(getargspec(func).args) - 1

        if attrs.get("default_values", None) is None:
            attrs["default_values"] = "None"
        else:
            # confirm number_of_elements == len(default_values)
            assert attrs["number_of_elements"] == _count(attrs["default_values"])

        # if repeat_command, set number_of_elements_per_command
        # if not set.
        if attrs.get("repeat_command", None) is not None and \
                attrs.get("number_of_elements_per_command", None) is None:
                    attrs["number_of_elements_per_command"] = len(getargspec(func).args) - 1
        return attrs

    @staticmethod
    def _update_proxyproperty_attrs(func, attrs):
        return _update_property_defaults(func, attrs)

    @staticmethod
    def xml(xmlstr):
        """Decorator that be used to directly add a ServerManager property XML for a method."""
        def generate(func, attrs):
            smproperty._append_xml(func, xmlstr)
        return _create_decorator(generate_xml_func=generate)

    @staticmethod
    def intvector(**kwargs):
        attrs = { "type" : "IntVectorProperty"}
        attrs.update(kwargs)
        return _create_decorator(attrs,
                update_func=smproperty._update_vectorproperty_defaults,
                generate_xml_func=smproperty._generate_xml)

    @staticmethod
    def doublevector(**kwargs):
        attrs = { "type" : "DoubleVectorProperty"}
        attrs.update(kwargs)
        return _create_decorator(attrs,
                update_func=smproperty._update_vectorproperty_defaults,
                generate_xml_func=smproperty._generate_xml)

    @staticmethod
    def idtypevector(**kwargs):
        attrs = { "type" : "IdTypeVectorProperty"}
        attrs.update(kwargs)
        return _create_decorator(attrs,
                update_func=smproperty._update_vectorproperty_defaults,
                generate_xml_func=smproperty._generate_xml)

    @staticmethod
    def stringvector(**kwargs):
        attrs = { "type" : "StringVectorProperty" }
        attrs.update(kwargs)
        return _create_decorator(attrs,
                update_func=smproperty._update_vectorproperty_defaults,
                generate_xml_func=smproperty._generate_xml)

    @staticmethod
    def proxy(**kwargs):
        attrs = { "type" : "ProxyProperty" }
        attrs.update(kwargs)
        return _create_decorator(attrs,
                update_func=smproperty._update_proxyproperty_attrs,
                generate_xml_func=smproperty._generate_xml)

    @staticmethod
    def input(**kwargs):
        attrs = { "type" : "InputProperty" }
        if kwargs.get("multiple_input", False) or kwargs.get("repeat_command", False):
            attrs["command"] = "AddInputConnection"
            # FIXME: input property doesn't support cleaning port connections alone :(
            attrs["clean_command"] = "RemoveAllInputs"
        else:
            attrs["command"] = "SetInputConnection"

        # todo: handle inputType
        attrs.update(kwargs)
        return _create_decorator(attrs,
                update_func=smproperty._update_property_defaults,
                generate_xml_func=smproperty._generate_xml)

    @staticmethod
    def dataarrayselection(name=None):
        def generate(func, attrs):
            xml="""<StringVectorProperty name="{name}Info"
                                         command="{command}"
                                         number_of_elements_per_command="2"
                                         information_only="1"
                                         si_class="vtkSIDataArraySelectionProperty" />
                   <StringVectorProperty name="{name}"
                                         information_property="{name}Info"
                                         command="{command}"
                                         number_of_elements_per_command="2"
                                         element_types="2 0"
                                         repeat_command="1"
                                         si_class="vtkSIDataArraySelectionProperty">
                     <ArraySelectionDomain name="array_list">
                        <RequiredProperties>
                            <Property function="ArrayList" name="{name}Info" />
                        </RequiredProperties>
                     </ArraySelectionDomain>
                   </StringVectorProperty>
            """.format(**attrs)
            smproperty._append_xml(func, xml)
        return _create_decorator({"name" : name},
                update_func=smproperty._update_property_defaults,
                generate_xml_func=generate)


class smdomain(object):
    """
    Provides decorators that add domains to properties.
    """
    @staticmethod
    def _append_xml(func, xml):
        domains = []
        if hasattr(func, "_pvsm_domain_xmls"):
            domains = func._pvsm_domain_xmls
        domains.append(xml)
        setattr(func, "_pvsm_domain_xmls", domains)

    @staticmethod
    def _generate_xml(func, attrs):
        smdomain._append_xml(func, _generate_xml(attrs, []))

    @staticmethod
    def xml(xmlstr):
        def generate(func, attrs):
            smdomain._append_xml(func, xmlstr)

        return _create_decorator({},
                generate_xml_func=generate)

    @staticmethod
    def doublerange(**kwargs):
        attrs = { "type" : "DoubleRangeDomain" , "name" : "range" }
        attrs.update(kwargs)
        return _create_decorator(attrs,
                generate_xml_func=smdomain._generate_xml)

    @staticmethod
    def intrange(**kwargs):
        attrs = { "type" : "IntRangeDomain" , "name" : "range" }
        attrs.update(kwargs)
        return _create_decorator(attrs,
                generate_xml_func=smdomain._generate_xml)

    @staticmethod
    def filelist(**kwargs):
        attrs = { "type" : "FileListDomain", "name" : "files" }
        attrs.update(kwargs)
        return _create_decorator(attrs,
                generate_xml_func=smdomain._generate_xml)

    @staticmethod
    def datatype(dataTypes, **kwargs):
        attrs = {"type" : "DataTypeDomain", "name": "input_type"}
        attrs.update(kwargs)
        def generate(func, attrs):
            type_xmls = []
            for atype in dataTypes:
                type_xmls.append(_generate_xml({"type" : "DataType", "value" : atype}, []))
            smdomain._append_xml(func, _generate_xml(attrs, type_xmls))

        return _create_decorator(attrs,
                generate_xml_func=generate)


class smhint(object):
    """Provides decorators that add hints to proxies and properties."""
    @staticmethod
    def _generate_xml(func, attrs):
        lhints = []
        if hasattr(func, "_pvsm_hints_xmls"):
            lhints = func._pvsm_hints_xmls
        lhints.append(_generate_xml(attrs, []))
        setattr(func, "_pvsm_hints_xmls", lhints)

    @staticmethod
    def xml(xmlstr):
        def generate(func, attrs):
            lhints = []
            if hasattr(func, "_pvsm_hints_xmls"):
                lhints = func._pvsm_hints_xmls
            lhints.append(xmlstr)
            setattr(func, "_pvsm_hints_xmls", lhints)
        return _create_decorator({},
                generate_xml_func=generate)

    @staticmethod
    def filechooser(extensions, file_description):
        attrs = {}
        attrs["type"] = "FileChooser"
        attrs["extensions"] = extensions
        attrs["file_description"] = file_description
        return _create_decorator(attrs,
                generate_xml_func=smhint._generate_xml)

def get_qualified_classname(classobj):
    if classobj.__module__ == "__main__":
        return classobj.__name__
    else:
        return "%s.%s" % (classobj.__module__, classobj.__name__)

class smproxy(object):
    """
    Provides decorators for class objects that should be exposed to
    ParaView.
    """
    @staticmethod
    def _update_proxy_defaults(classobj, attrs):
        if attrs.get("name", None) is None:
            attrs["name"] = classobj.__name__
        if attrs.get("class", None) is None:
            attrs["class"] = get_qualified_classname(classobj)
        if attrs.get("label", None) is None:
            attrs["label"] = attrs["name"]
        return attrs

    @staticmethod
    def _generate_xml(classobj, attrs):
        nested_xmls = []

        classobj = _undecorate(classobj)
        if hasattr(classobj, "_pvsm_property_xmls"):
            val = getattr(classobj, "_pvsm_property_xmls")
            if type(val) == type([]):
                nested_xmls += val
            else:
                nested_xmls.append(val)

        prop_xmls_dict = {}
        for pname, val in classobj.__dict__.items():
            val = _undecorate(val)
            if callable(val) and hasattr(val, "_pvsm_property_xmls"):
                pxmls = getattr(val, "_pvsm_property_xmls")
                if len(pxmls) > 1:
                    raise RuntimeError("Multiple property definitions on the same"\
                            "method are not supported.")
                prop_xmls_dict[pname] = pxmls[0]


        # since the order of the properties keeps on changing between invocations,
        # let's sort them by the name for consistency. In future, we may put in
        # extra logic to preserve order they were defined in the class
        nested_xmls += [prop_xmls_dict[key] for key in sorted(prop_xmls_dict.keys())]

        if attrs.get("support_reload", True):
            nested_xmls.insert(0, """
                <Property name="Reload Python Module" panel_widget="command_button">
                    <Documentation>Reload the Python module.</Documentation>
                </Property>""")

        if hasattr(classobj, "_pvsm_hints_xmls"):
            hints = [h  for h in classobj._pvsm_hints_xmls]
            nested_xmls.append(_generate_xml({"type":"Hints"}, hints))

        proxyxml = _generate_xml(attrs, nested_xmls)
        groupxml = _generate_xml({"type":"ProxyGroup", "name":attrs.get("group")},
                                [proxyxml])
        smconfig = _generate_xml({"type":"ServerManagerConfiguration"}, [groupxml])
        setattr(classobj, "_pvsm_proxy_xml", smconfig)

    @staticmethod
    def source(**kwargs):
        attrs = {}
        attrs["type"] = "SourceProxy"
        attrs["group"] = "sources"
        attrs["si_class"] = "vtkSIPythonSourceProxy"
        attrs.update(kwargs)
        return _create_decorator(attrs,
                update_func=smproxy._update_proxy_defaults,
                generate_xml_func=smproxy._generate_xml)

    @staticmethod
    def filter(**kwargs):
        attrs = { "group" : "filters" }
        attrs.update(kwargs)
        return smproxy.source(**attrs)

    @staticmethod
    def reader(file_description, extensions=None, filename_patterns=None, is_directory=False, **kwargs):
        """
        Decorates a reader. Either `filename_patterns` or `extensions` must be
        provided.
        """
        if extensions is None and filename_patterns is None:
            raise RuntimeError("Either `filename_patterns` or `extensions` must be provided for a reader.")

        attrs = { "type" : "ReaderFactory" }
        attrs["file_description"] = file_description
        attrs["extensions"] = extensions
        attrs["filename_patterns"] = filename_patterns
        attrs["is_directory"] = "1" if is_directory else None
        _xml = _generate_xml(attrs, [])
        def decorator(func):
            f = smhint.xml(_xml)(func)
            return smproxy.source(**kwargs)(f)
        return decorator

    @staticmethod
    def writer(file_description, extensions, **kwargs):
        """
        Decorates a writer.
        """
        assert file_description is not None and extensions is not None

        attrs = { "type" : "WriterFactory" }
        attrs["file_description"] = file_description
        attrs["extensions"] = extensions
        _xml = _generate_xml(attrs, [])
        def decorator(func):
            f = smhint.xml(_xml)(func)
            return smproxy.source(group="writers", type="WriterProxy", **kwargs)(f)
        return decorator


def get_plugin_xmls(module_or_package):
    """helper function called by vtkPVPythonAlgorithmPlugin to discover
    all "proxy" decorated classes in the module or package. We don't recurse
    into the package, on simply needs to export all classes that form the
    ParaView plugin in the __init__.py for the package."""
    from inspect import ismodule, isclass
    items = []
    if ismodule(module_or_package):
        items = module_or_package.__dict__.items()
    elif hasattr(module_or_package, "items"):
        items = module_or_package.items()

    xmls = []
    for (k,v) in items:
        v = _undecorate(v)
        if hasattr(v, "_pvsm_proxy_xml"):
            xmls.append(getattr(v, "_pvsm_proxy_xml"))
    return xmls

def get_plugin_name(module_or_package):
    """helper function called by vtkPVPythonAlgorithmPlugin to discover
    ParaView plugin name, if any."""
    from inspect import ismodule, isclass
    if ismodule(module_or_package) and hasattr(module_or_package, "paraview_plugin_name"):
        return str(getattr(module_or_package, "paraview_plugin_name"))
    else:
        return module_or_package.__name__

def get_plugin_version(module_or_package):
    """helper function called by vtkPVPythonAlgorithmPlugin to discover
    ParaView plugin version, if any."""
    from inspect import ismodule, isclass
    if ismodule(module_or_package) and hasattr(module_or_package, "paraview_plugin_version"):
        return str(getattr(module_or_package, "paraview_plugin_version"))
    else:
        return "(unknown)"

def load_plugin(filepath, default_modulename=None):
    """helper function called by vtkPVPythonAlgorithmPlugin to load
    a python file."""

    # should we scope these under a plugins namespace?
    if default_modulename:
        modulename = default_modulename
    else:
        import os.path
        modulename = "%s" % os.path.splitext(os.path.basename(filepath))[0]

    try:
        # for Python 3.4+
        from importlib.util import spec_from_file_location, module_from_spec
        spec = spec_from_file_location(modulename, filepath)
        module = module_from_spec(spec)
        spec.loader.exec_module(module)
    except ImportError:
        # for Python 2.7
        import imp
        module = imp.load_source(modulename, filepath)

    import sys
    sys.modules[modulename] = module
    return module

def reload_plugin_module(module):
    """helper function to reload a plugin module previously loaded via
    load_plugin"""
    from inspect import getsourcefile, ismodule
    if ismodule(module) and getsourcefile(module):
        return load_plugin(getsourcefile(module), module.__name__)
    return module



import paraview


class ProgressObject():
    '''Dummy progress reporter. See git history for how this used to work.'''
    def __init__(self):
        pass

    def UpdateProgress(self, progress):
        ''' 'progress' is expected to be in the range [0, 1]. '''
        pass

    def StartEvent(self):
        pass

    def EndEvent(self):
        pass



"""
Module that uses one of the available back end libraries to write out image
files for cinema's file store class.
"""

import numpy
import zlib
import os
import warnings

exrEnabled = False
try:
    import oexr_helper as exr
    exrEnabled = True
    print(
        "Imported OpenEXR, will default to *.exr for float images (z-buffer, "
        "value, etc.).")
except ImportError:
    pass

pilEnabled = False
try:
    import PIL.Image
    import PIL.ImImagePlugin
    pilEnabled = True
except ImportError:
    pass

vtkEnabled = False
try:
    import sys
    if "paraview" in sys.modules:
        from paraview.vtk.vtkIOImage import (vtkPNGReader,
                                             vtkBMPReader,
                                             vtkPNMReader,
                                             vtkTIFFReader,
                                             vtkJPEGReader,
                                             vtkPNGWriter,
                                             vtkBMPWriter,
                                             vtkPNMWriter,
                                             vtkTIFFWriter,
                                             vtkJPEGWriter)
        from paraview.vtk.vtkIOAsynchronous import vtkThreadedImageWriter
        from paraview.vtk.vtkCommonDataModel import vtkImageData
        from paraview import numpy_support as n2v
    else:
        from vtk import (vtkPNGReader,
                         vtkBMPReader,
                         vtkPNMReader,
                         vtkTIFFReader,
                         vtkJPEGReader,
                         vtkPNGWriter,
                         vtkBMPWriter,
                         vtkPNMWriter,
                         vtkTIFFWriter,
                         vtkJPEGWriter,
                         vtkImageData)
        from vtk.vtkIOAsynchronous import vtkThreadedImageWriter
        from vtk.util import numpy_support as n2v
    vtkEnabled = True
except ImportError:
    pass


class RasterWrangler(object):
    """
    Isolates the specifics of raster file formats from the cinema store.
    In particular this delegates the task to one or more subsidiary modules.
    The choice of which is open to facilitate packaging in different
    places, i.e. PIL for desktop and small packages, VTK for HPC contexts.
    """

    def __init__(self):
        self.threadedwriter = None
        self.backends = set()
        if exrEnabled:
            self.backends.add("OpenEXR")
        elif vtkEnabled:
            self.backends.add("VTK")
            self.threadedwriter = vtkThreadedImageWriter()
            self.threadedwriter.Initialize()
        elif pilEnabled:
            self.backends.add("PIL")

        # self.dontCompressFloatVals = False  # don't expect we'll need this
        # self.dontConvertValsToFloat = False  # nor this

    def enableOpenEXR(self):
        """Try to turn on OpenEXR file IO support"""
        if exrEnabled:
            self.backends.add("OpenEXR")
        else:
            warnings.warn("OpenEXR module not found", ImportWarning)

    def enablePIL(self):
        """Try to turn on PIL file IO support"""
        if pilEnabled:
            self.backends.add("PIL")
        else:
            warnings.warn("PIL module not found", ImportWarning)

    def enableVTK(self):
        """Try to turn on VTK file IO support"""
        if vtkEnabled:
            self.backends.add("VTK")
        else:
            warnings.warn("VTK module not found", ImportWarning)

    def _make_writer(self, filename):
        "Internal function."
        extension = None
        parts = filename.split('.')
        if len(parts) > 1:
            extension = parts[-1]
        else:
            raise RuntimeError(
                "Filename has no extension, cannot guess writer to use")

        if extension == 'png':
            return vtkPNGWriter()
        elif extension == 'bmp':
            return vtkBMPWriter()
        elif extension == 'ppm':
            return vtkPNMWriter()
        elif extension == 'tif' or extension == 'tiff':
            return vtkTIFFWriter()
        elif extension == 'jpg' or extension == 'jpeg':
            return vtkJPEGWriter()
        elif extension == 'vti':
            return vtkXMLImageDataWriter()  # noqa: F821
        else:
            raise RuntimeError(
                "Cannot infer filetype from extension:", extension)

    def _make_reader(self, filename):
        "Internal function."
        extension = None
        parts = filename.split('.')
        if len(parts) > 1:
            extension = parts[-1]
        else:
            raise RuntimeError(
                "Filename has no extension, please guess reader to use")

        if extension == 'png':
            return vtkPNGReader()
        elif extension == 'bmp':
            return vtkBMPReader()
        elif extension == 'ppm':
            return vtkPNMReader()
        elif extension == 'tif' or extension == 'tiff':
            return vtkTIFFReader()
        elif extension == 'jpg' or extension == 'jpeg':
            return vtkJPEGReader()
        elif extension == 'vti':
            return vtkXMLImageDataReader()  # noqa: F821
        else:
            raise RuntimeError(
                "Cannot infer filetype from extension:", extension)

    def genericreader(self, fname):
        """read generic binary data dump"""
        with open(fname, "r") as file:
            return file.read()

    def genericwriter(self, imageslice, fname):
        """write generic binary data dump"""
        if self.threadedwriter is not None:
            height = imageslice.shape[1]
            width = imageslice.shape[0]
            contig = imageslice.reshape(height*width)
            vtkarray = n2v.numpy_to_vtk(contig)
            id = vtkImageData()
            id.SetExtent(0, height-1, 0, width-1, 0, 0)
            id.GetPointData().SetScalars(vtkarray)
            self.threadedwriter.EncodeAndWrite(id, fname)
        else:
            with open(fname, "w") as file:
                file.write(imageslice)
        return fname

    def rgbreader(self, fname):
        """opens a color image file and returns it as a color buffer"""
        if "VTK" in self.backends:
            reader = self._make_reader(fname)
            reader.SetFileName(fname)
            reader.Update()
            id = reader.GetOutput()
            ext = id.GetExtent()
            width = ext[1]+1
            height = ext[3]+1
            imageslice = n2v.vtk_to_numpy(id.GetPointData().GetArray(0))
            imageslice = imageslice.reshape(height, width, 3)
            return numpy.flipud(imageslice)
        elif "PIL" in self.backends:
            try:
                im = PIL.Image.open(fname)
                # this image could be any format, but we require RGB
                # as we handle this image
                conv_im = im.convert("RGB")

                return numpy.array(
                    conv_im, numpy.uint8).reshape(im.size[1], im.size[0], 3)
            except:
                return None

        else:
            print("Warning: need PIL or VTK to read from " + fname)

    def rgbwriter(self, imageslice, fname):
        """takes in a color buffer and writes it as an image file"""
        if "VTK" in self.backends:
            height = imageslice.shape[1]
            width = imageslice.shape[0]
            contig = imageslice.reshape(height*width, 3)
            vtkarray = n2v.numpy_to_vtk(contig)
            id = vtkImageData()
            id.SetExtent(0, height-1, 0, width-1, 0, 0)
            id.GetPointData().SetScalars(vtkarray)

            if self.threadedwriter is not None:
                self.threadedwriter.EncodeAndWrite(id, fname)
            else:
                writer = self._make_writer(fname)
                writer.SetInputData(id)
                writer.SetFileName(fname)
                writer.Write()
            return fname

        elif "PIL" in self.backends:
            imageslice = numpy.flipud(imageslice)
            pimg = PIL.Image.fromarray(imageslice)
            pimg.save(fname)
            return fname

        else:
            raise ValueError("Warning: need PIL or VTK to write to " + fname)

    def valuewriter(self, imageSlice, fname, vrange):
        """ Takes in either a (1C) float or a RGB (3C) buffer and writes it as
        an image file."""

        # Adjust the filename, replace png with .im
        baseName, ext = os.path.splitext(fname)
        adjustedName = baseName + self.floatExtension()

        dimensions = imageSlice.shape
        if len(dimensions) == 2 and imageSlice.dtype == numpy.float32:
            # Given as single channel floating point buffer.
            return self.zwriter(imageSlice, adjustedName)

        elif (len(dimensions) > 2) and (dimensions[2] == 3):
            # if self.dontConvertValsToFloat
            #     self.rgbwriter(imageSlice, fname)
            #     return

            # Given an RGB buffer
            # Convert it to a floating point buffer for consistency
            # TODO: just one copy of this code in cinema
            w0 = numpy.left_shift(imageSlice[:, :, 0].astype(numpy.uint32), 16)
            w1 = numpy.left_shift(imageSlice[:, :, 1].astype(numpy.uint32), 8)
            w2 = imageSlice[:, :, 2].astype(numpy.uint32)
            value = numpy.bitwise_or(w0, w1)
            value = numpy.bitwise_or(value, w2)
            value = numpy.subtract(value, 1)
            value = value.astype(numpy.float32)
            adjusted_val = numpy.divide(value, float(0xFFFFFE))

            return self.zwriter(adjusted_val, adjustedName)

        else:
            raise ValueError("Invalid dimensions for a value raster.")


    def valuereader(self, fname, shape=None):
        """ Opens a value image file and returns it as either a color buffer
        or a floating point array (depending on how the image was exported)."""
        baseName, ext = os.path.splitext(fname)
        if ext == self.floatExtension():
            # Treat as single channel floating point buffer.
            return self.zreader(fname, shape)
        else:
            # Treat as a RGB buffer
            return self.rgbreader(fname)

    def floatExtension(self):
        """determine file extension for depth images"""
        if "OpenEXR" in self.backends:
            return ".exr"

        else:
            return ".im"

    def zreader(self, fname, shape=None):
        """reads a depth file to make a depth buffer"""

        if "OpenEXR" in self.backends:
            return exr.load_depth(fname)

        if "PIL" in self.backends:
            # for backwards compatibility with Bacall, remove a.s.a.p
            try:
                im = PIL.Image.open(fname)
                return numpy.array(
                    im, numpy.float32).reshape(im.size[1], im.size[0])
            except:
                pass

        # Try to open as zlib compressed files first
        baseName, ext = os.path.splitext(fname)
        try:
            # Adjust the filename, replace .im with .Z
            adjustedName = baseName + ".Z"

            with open(adjustedName, mode='rb') as file:
                compresseddata = file.read()
            flatarr = numpy.fromstring(zlib.decompress(compresseddata),
                                       numpy.float32)

            if not shape:
                shape = flatarr.shape

            imageslice = flatarr.reshape(shape)
        # Fall back on numpy compressed format
        except IOError:
            # Adjust the filename, replace .im with .npz
            adjustedName = baseName + ".npz"

            with open(adjustedName, mode='rb') as file:
                with numpy.load(file) as tz:
                    imageslice = tz[tz.files[0]]
        return imageslice

    def zwriter(self, imageslice, fname):
        """takes in a depth buffer and writes it as a depth file"""

        if "OpenEXR" in self.backends:
            imageslice = numpy.flipud(imageslice)
            exr.save_depth(imageslice, fname)
            return fname

        # if self.dontCompressFloatVals:
        #     if "VTK" in self.backends:
        #         height = imageslice.shape[1]
        #         width = imageslice.shape[0]
        #
        #         file = open(fname, mode='w')
        #         file.write("Image type: L 32F image\r\n")
        #         file.write("Name: A cinema depth image\r\n")
        #         file.write(
        #             "Image size (x*y): "+str(height) +
        #             "*" + str(width) + "\r\n")
        #         file.write("File size (no of images): 1\r\n")
        #         file.write(chr(26))
        #         imageslice.tofile(file)
        #         file.close()
        #         return
        #
        #     imageslice = numpy.flipud(imageslice)
        #     pimg = PIL.Image.fromarray(imageslice)
        #     # TODO:
        #     # don't let ImImagePlugin.py insert the Name: filename in
        #     line two. why? because ImImagePlugin.py reader has a 100
        #     character limit
        #     pimg.save(fname)

        imageslice = numpy.flipud(imageslice)
        # Adjust the filename, replace .im with .npz
        baseName, ext = os.path.splitext(fname)
        adjustedName = baseName + ".Z"

        if self.threadedwriter is not None:
            height = imageslice.shape[1]
            width = imageslice.shape[0]
            contig = imageslice.reshape(height*width)
            vtkarray = n2v.numpy_to_vtk(contig)
            id = vtkImageData()
            id.SetExtent(0, height-1, 0, width-1, 0, 0)
            id.GetPointData().SetScalars(vtkarray)
            self.threadedwriter.EncodeAndWrite(id, adjustedName)
        else:
            with open(adjustedName, mode='wb') as file:
                file.write(zlib.compress(numpy.array(imageslice)))
        return adjustedName

    def assertvalidimage(self, filename):
        """tests that a given file is syntactically correct"""

        # Try .Z form
        baseName, ext = os.path.splitext(filename)
        adjustedName = baseName + ".Z"
        if os.path.isfile(adjustedName):
            with open(adjustedName, mode='rb') as file:
                compresseddata = file.read()

            imageslice = numpy.fromstring(zlib.decompress(compresseddata),
                                          numpy.float32)
            if not isinstance(imageslice, numpy.ndarray):
                raise IOError(adjustedName + " does not load correctly.")
            return

        # Try .npz form
        adjustedName = baseName + ".npz"
        if os.path.isfile(adjustedName):
            with open(adjustedName, mode='rb') as file:
                with numpy.load(file) as tz:
                    if not isinstance(tz[tz.files[0]], numpy.ndarray):
                        raise IOError(
                            adjustedName + " does not load correctly.")
            return

        if not os.path.isfile(filename):
            raise IOError(filename + " does not exist.")

        if "OpenEXR" in self.backends:
            if not exr.isOpenExrFile(filename):
                raise IOError(filename + " cannot be opened using OpenEXR.")

        elif "VTK" in self.backends:
            reader = self._make_reader(filename)
            if not reader.CanReadFile(filename):
                raise IOError("VTK Cannot open file " + filename)

        elif "PIL" in self.backends:
            try:
                PIL.Image.open(filename)
            except IOError:
                raise

        else:
            raise RuntimeError(
                "Warning: need OpenEXR or PIL or VTK to validate file store")



"""Miscellaneous functions and classes that don't fit into specific
categories."""

import sys, os

def calldata_type(type):
    """set_call_data_type(type) -- convenience decorator to easily set the CallDataType attribute
    for python function used as observer callback.
    For example:

    import vtkmodules.util.calldata_type
    import vtkmodules.util.vtkConstants
    import vtkmodules.vtkCommonCore import vtkCommand, vtkLookupTable

    @calldata_type(vtkConstants.VTK_STRING)
    def onError(caller, event, calldata):
        print("caller: %s - event: %s - msg: %s" % (caller.GetClassName(), event, calldata))

    lt = vtkLookupTable()
    lt.AddObserver(vtkCommand.ErrorEvent, onError)
    lt.SetTableRange(2,1)
    """
    from vtkmodules import vtkCommonCore
    supported_call_data_types = ['string0', vtkCommonCore.VTK_STRING,
            vtkCommonCore.VTK_OBJECT, vtkCommonCore.VTK_INT,
            vtkCommonCore.VTK_LONG, vtkCommonCore.VTK_DOUBLE, vtkCommonCore.VTK_FLOAT]

    if type not in supported_call_data_types:
        raise TypeError("'%s' is not a supported VTK call data type. Supported types are: %s" % (type, supported_call_data_types))

    def wrap(f):
        f.CallDataType = type
        return f

    return wrap

#----------------------------------------------------------------------
# the following functions are for the vtk regression testing and examples

def vtkGetDataRoot():
    """vtkGetDataRoot() -- return vtk example data directory"""
    dataRoot = None
    for i, argv in enumerate(sys.argv):
        if argv == '-D' and i+1 < len(sys.argv):
            dataRoot = sys.argv[i+1]

    if dataRoot is None:
        dataRoot = os.environ.get('VTK_DATA_ROOT', '../../../../VTKData')

    return dataRoot

def vtkGetTempDir():
    """vtkGetTempDir() -- return vtk testing temp dir"""
    tempDir = None
    for i, argv in enumerate(sys.argv):
        if argv == '-T' and i+1 < len(sys.argv):
            tempDir = sys.argv[i+1]

    if tempDir is None:
        tempDir = '.'

    return tempDir

def vtkRegressionTestImage(renWin):
    """vtkRegressionTestImage(renWin) -- produce regression image for window

    This function writes out a regression .png file for a vtkWindow.
    Does anyone involved in testing care to elaborate?
    """
    from vtkmodules.vtkRenderingCore import vtkWindowToImageFilter
    from vtkmodules.vtkIOImage import vtkPNGReader
    from vtkmodules.vtkImagingCore import vtkImageDifference

    fname = None
    for i, argv in enumerate(sys.argv):
        if argv == '-V' and i+1 < len(sys.argv):
            fname = os.path.join(vtkGetDataRoot(), sys.argv[i+1])

    if fname is None:
        return 2

    else:
        rt_w2if = vtkWindowToImageFilter()
        rt_w2if.SetInput(renWin)

        if not os.path.isfile(fname):
            rt_pngw = vtkPNGWriter()
            rt_pngw.SetFileName(fname)
            rt_pngw.SetInputConnection(rt_w2if.GetOutputPort())
            rt_pngw.Write()
            rt_pngw = None

        rt_png = vtkPNGReader()
        rt_png.SetFileName(fname)

        rt_id = vtkImageDifference()
        rt_id.SetInputConnection(rt_w2if.GetOutputPort())
        rt_id.SetImageConnection(rt_png.GetOutputPort())
        rt_id.Update()

        if rt_id.GetThresholdedError() <= 10:
            return 1
        else:
            sys.stderr.write('Failed image test: %f\n'
                             % rt_id.GetThresholdedError())
            return 0



from vtk import *
from vtk.web import getJSArrayType
from vtk.web.camera import *
from vtk.web.query_data_model import *
from vtk.web import iteritems, buffer
import json, os, math, gzip, shutil

# Global helper variables
encode_codes = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'

# -----------------------------------------------------------------------------
# Capture image from render window
# -----------------------------------------------------------------------------

class CaptureRenderWindow(object):
    def __init__(self, magnification=1):
        self.windowToImage = vtkWindowToImageFilter()
        self.windowToImage.SetScale(magnification)
        self.windowToImage.SetInputBufferTypeToRGB()
        self.windowToImage.ReadFrontBufferOn()
        self.writer = None

    def SetRenderWindow(self, renderWindow):
        self.windowToImage.SetInput(renderWindow)

    def SetFormat(self, mimeType):
        if mimeType == 'image/png':
            self.writer = vtkPNGWriter()
            self.writer.SetInputConnection(self.windowToImage.GetOutputPort())
        elif mimeType == 'image/jpg':
            self.writer = vtkJPEGWriter()
            self.writer.SetInputConnection(self.windowToImage.GetOutputPort())

    def writeImage(self, path):
        if self.writer:
            self.windowToImage.Modified()
            self.windowToImage.Update()
            self.writer.SetFileName(path)
            self.writer.Write()

# -----------------------------------------------------------------------------
# Basic Dataset Builder
# -----------------------------------------------------------------------------

class DataSetBuilder(object):
    def __init__(self, location, camera_data, metadata={}, sections={}):
        self.dataHandler = DataHandler(location)
        self.cameraDescription = camera_data
        self.camera = None
        self.imageCapture = CaptureRenderWindow()

        for key, value in iteritems(metadata):
            self.dataHandler.addMetaData(key, value)

        for key, value in iteritems(sections):
            self.dataHandler.addSection(key, value)

    def getDataHandler(self):
        return self.dataHandler

    def getCamera(self):
        return self.camera

    def updateCamera(self, camera):
        update_camera(self.renderer, camera)
        self.renderWindow.Render()

    def start(self, renderWindow=None, renderer=None):
        if renderWindow:
            # Keep track of renderWindow and renderer
            self.renderWindow = renderWindow
            self.renderer = renderer

            # Initialize image capture
            self.imageCapture.SetRenderWindow(renderWindow)

            # Handle camera if any
            if self.cameraDescription:
                if self.cameraDescription['type'] == 'spherical':
                    self.camera = create_spherical_camera(renderer, self.dataHandler, self.cameraDescription['phi'], self.cameraDescription['theta'])
                elif self.cameraDescription['type'] == 'cylindrical':
                    self.camera = create_cylindrical_camera(renderer, self.dataHandler, self.cameraDescription['phi'], self.cameraDescription['translation'])

            # Update background color
            bgColor = renderer.GetBackground()
            bgColorString = 'rgb(%d, %d, %d)' % tuple(int(bgColor[i]*255) for i in range(3))
            self.dataHandler.addMetaData('backgroundColor', bgColorString)

        # Update file patterns
        self.dataHandler.updateBasePattern()

    def stop(self):
        self.dataHandler.writeDataDescriptor()

# -----------------------------------------------------------------------------
# Image Dataset Builder
# -----------------------------------------------------------------------------

class ImageDataSetBuilder(DataSetBuilder):
    def __init__(self, location, imageMimeType, cameraInfo, metadata={}, sections={}):
        DataSetBuilder.__init__(self, location, cameraInfo, metadata, sections)
        imageExtenstion = '.' + imageMimeType.split('/')[1]
        self.dataHandler.registerData(name='image', type='blob', mimeType=imageMimeType, fileName=imageExtenstion)
        self.imageCapture.SetFormat(imageMimeType)

    def writeImage(self):
        self.imageCapture.writeImage(self.dataHandler.getDataAbsoluteFilePath('image'))

    def writeImages(self):
        for cam in self.camera:
            update_camera(self.renderer, cam)
            self.renderWindow.Render()
            self.imageCapture.writeImage(self.dataHandler.getDataAbsoluteFilePath('image'))

# -----------------------------------------------------------------------------
# Volume Composite Dataset Builder
# -----------------------------------------------------------------------------
class VolumeCompositeDataSetBuilder(DataSetBuilder):
    def __init__(self, location, imageMimeType, cameraInfo, metadata={}, sections={}):
        DataSetBuilder.__init__(self, location, cameraInfo, metadata, sections)

        self.dataHandler.addTypes('volume-composite', 'rgba+depth')

        self.imageMimeType = imageMimeType
        self.imageExtenstion = '.' + imageMimeType.split('/')[1]

        if imageMimeType == 'image/png':
            self.imageWriter = vtkPNGWriter()
        if imageMimeType == 'image/jpg':
            self.imageWriter = vtkJPEGWriter()

        self.imageDataColor = vtkImageData()
        self.imageWriter.SetInputData(self.imageDataColor)

        self.imageDataDepth = vtkImageData()
        self.depthToWrite = None

        self.layerInfo = {}
        self.colorByMapping = {}
        self.compositePipeline = {'layers': [], 'dimensions': [], 'fields': {}, 'layer_fields': {}, 'pipeline': []}
        self.activeDepthKey = ''
        self.activeRGBKey = ''
        self.nodeWithChildren = {}

    def _getColorCode(self, colorBy):
        if colorBy in self.colorByMapping:
            # The color code exist
            return self.colorByMapping[colorBy]
        else:
            # No color code assigned yet
            colorCode = encode_codes[len(self.colorByMapping)]
            # Assign color code
            self.colorByMapping[colorBy] = colorCode
            # Register color code with color by value
            self.compositePipeline['fields'][colorCode] = colorBy
            # Return the color code
            return colorCode

    def _getLayerCode(self, parent, layerName):
        if layerName in self.layerInfo:
            # Layer already exist
            return (self.layerInfo[layerName]['code'], False)
        else:
            layerCode = encode_codes[len(self.layerInfo)]
            self.layerInfo[layerName] = { 'code': layerCode, 'name': layerName, 'parent': parent }
            self.compositePipeline['layers'].append(layerCode)
            self.compositePipeline['layer_fields'][layerCode] = []

            # Let's register it in the pipeline
            if parent:
                if parent not in self.nodeWithChildren:
                    # Need to create parent
                    rootNode = {'name': parent, 'ids': [], 'children':[]}
                    self.nodeWithChildren[parent] = rootNode
                    self.compositePipeline['pipeline'].append(rootNode)

                # Add node to its parent
                self.nodeWithChildren[parent]['children'].append({'name': layerName, 'ids': [layerCode]})
                self.nodeWithChildren[parent]['ids'].append(layerCode)

            else:
                self.compositePipeline['pipeline'].append({'name': layerName, 'ids': [layerCode]})

            return (layerCode, True)

    def _needToRegisterColor(self, layerCode, colorCode):
        if colorCode in self.compositePipeline['layer_fields'][layerCode]:
            return False
        else:
            self.compositePipeline['layer_fields'][layerCode].append(colorCode)
            return True

    def activateLayer(self, parent, name, colorBy):
        layerCode, needToRegisterDepth = self._getLayerCode(parent, name)
        colorCode = self._getColorCode(colorBy)
        needToRegisterColor = self._needToRegisterColor(layerCode, colorCode)

        # Update active keys
        self.activeDepthKey = '%s_depth' % layerCode
        self.activeRGBKey   = '%s%s_rgb' % (layerCode, colorCode)

        # Need to register data
        if needToRegisterDepth:
            self.dataHandler.registerData(name=self.activeDepthKey, type='array', fileName='/%s_depth.uint8' % layerCode, categories=[ layerCode ])

        if needToRegisterColor:
            self.dataHandler.registerData(name=self.activeRGBKey, type='blob', fileName='/%s%s_rgb%s' % (layerCode, colorCode, self.imageExtenstion), categories=[ '%s%s' % (layerCode, colorCode) ], mimeType=self.imageMimeType)

    def writeData(self, mapper):
        width = self.renderWindow.GetSize()[0]
        height = self.renderWindow.GetSize()[1]

        if not self.depthToWrite:
            self.depthToWrite = bytearray(width * height)

        for cam in self.camera:
            self.updateCamera(cam)
            imagePath = self.dataHandler.getDataAbsoluteFilePath(self.activeRGBKey)
            depthPath = self.dataHandler.getDataAbsoluteFilePath(self.activeDepthKey)

            # -----------------------------------------------------------------
            # Write Image
            # -----------------------------------------------------------------
            mapper.GetColorImage(self.imageDataColor)
            self.imageWriter.SetFileName(imagePath)
            self.imageWriter.Write()

            # -----------------------------------------------------------------
            # Write Depth
            # -----------------------------------------------------------------
            mapper.GetDepthImage(self.imageDataDepth)
            inputArray = self.imageDataDepth.GetPointData().GetArray(0)
            size = inputArray.GetNumberOfTuples()
            for idx in range(size):
                self.depthToWrite[idx] = int(inputArray.GetValue(idx))

            with open(depthPath, 'wb') as f:
                f.write(self.depthToWrite)

    def start(self, renderWindow, renderer):
        DataSetBuilder.start(self, renderWindow, renderer)
        self.camera.updatePriority([2,1])

    def stop(self, compress=True):
        # Push metadata
        self.compositePipeline['dimensions'] = self.renderWindow.GetSize()
        self.compositePipeline['default_pipeline'] = 'A'.join(self.compositePipeline['layers']) + 'A'
        self.dataHandler.addSection('CompositePipeline', self.compositePipeline)

        # Write metadata
        DataSetBuilder.stop(self)

        if compress:
            for root, dirs, files in os.walk(self.dataHandler.getBasePath()):
                print('Compress', root)
                for name in files:
                    if '.uint8' in name and '.gz' not in name:
                        with open(os.path.join(root, name), 'rb') as f_in:
                            with gzip.open(os.path.join(root, name + '.gz'), 'wb') as f_out:
                                shutil.copyfileobj(f_in, f_out)
                        os.remove(os.path.join(root, name))

# -----------------------------------------------------------------------------
# Data Prober Dataset Builder
# -----------------------------------------------------------------------------
class DataProberDataSetBuilder(DataSetBuilder):
    def __init__(self, location, sampling_dimesions, fields_to_keep, custom_probing_bounds = None, metadata={}):
        DataSetBuilder.__init__(self, location, None, metadata)
        self.fieldsToWrite = fields_to_keep
        self.resamplerFilter = vtkPResampleFilter()
        self.resamplerFilter.SetSamplingDimension(sampling_dimesions)
        if custom_probing_bounds:
            self.resamplerFilter.SetUseInputBounds(0)
            self.resamplerFilter.SetCustomSamplingBounds(custom_probing_bounds)
        else:
            self.resamplerFilter.SetUseInputBounds(1)

        # Register all fields
        self.dataHandler.addTypes('data-prober', 'binary')
        self.DataProber = { 'types': {}, 'dimensions': sampling_dimesions, 'ranges': {}, 'spacing': [1,1,1] }
        for field in self.fieldsToWrite:
            self.dataHandler.registerData(name=field, type='array', fileName='/%s.array' % field)

    def setDataToProbe(self, dataset):
        self.resamplerFilter.SetInputData(dataset)

    def setSourceToProbe(self, source):
        self.resamplerFilter.SetInputConnection(source.GetOutputPort())

    def writeData(self):
        self.resamplerFilter.Update()
        arrays = self.resamplerFilter.GetOutput().GetPointData()
        for field in self.fieldsToWrite:
            array = arrays.GetArray(field)
            if array:
                b = buffer(array)
                with open(self.dataHandler.getDataAbsoluteFilePath(field), 'wb') as f:
                    f.write(b)

                self.DataProber['types'][field] = getJSArrayType(array)
                if field in self.DataProber['ranges']:
                    dataRange = array.GetRange()
                    if dataRange[0] < self.DataProber['ranges'][field][0]:
                        self.DataProber['ranges'][field][0] = dataRange[0]
                    if dataRange[1] > self.DataProber['ranges'][field][1]:
                        self.DataProber['ranges'][field][1] = dataRange[1]
                else:
                    self.DataProber['ranges'][field] = [array.GetRange()[0], array.GetRange()[1]]
            else:
                print('No array for', field)
                print(self.resamplerFilter.GetOutput())

    def stop(self, compress=True):
        # Push metadata
        self.dataHandler.addSection('DataProber', self.DataProber)

        # Write metadata
        DataSetBuilder.stop(self)

        if compress:
            for root, dirs, files in os.walk(self.dataHandler.getBasePath()):
                print('Compress', root)
                for name in files:
                    if '.array' in name and '.gz' not in name:
                        with open(os.path.join(root, name), 'rb') as f_in:
                            with gzip.open(os.path.join(root, name + '.gz'), 'wb') as f_out:
                                shutil.copyfileobj(f_in, f_out)
                        os.remove(os.path.join(root, name))

# -----------------------------------------------------------------------------
# Sorted Composite Dataset Builder
# -----------------------------------------------------------------------------
class ConvertVolumeStackToSortedStack(object):
    def __init__(self, width, height):
        self.width = width
        self.height = height
        self.layers = 0

    def convert(self, directory):
        imagePaths = {}
        depthPaths = {}
        layerNames = []
        for fileName in os.listdir(directory):
            if '_rgb' in fileName or '_depth' in fileName:
                fileId = fileName.split('_')[0][0]
                if '_rgb' in fileName:
                    imagePaths[fileId] = os.path.join(directory, fileName)
                else:
                    layerNames.append(fileId)
                    depthPaths[fileId] = os.path.join(directory, fileName)

        layerNames.sort()

        if len(layerNames) == 0:
            return

        # Load data in Memory
        depthArrays = []
        imageReader = vtkPNGReader()
        numberOfValues = self.width * self.height * len(layerNames)
        imageSize = self.width * self.height
        self.layers = len(layerNames)

        # Write all images as single buffer
        opacity = vtkUnsignedCharArray()
        opacity.SetNumberOfComponents(1)
        opacity.SetNumberOfTuples(numberOfValues)

        intensity = vtkUnsignedCharArray()
        intensity.SetNumberOfComponents(1)
        intensity.SetNumberOfTuples(numberOfValues)

        for layer in range(self.layers):
            imageReader.SetFileName(imagePaths[layerNames[layer]])
            imageReader.Update()

            rgbaArray = imageReader.GetOutput().GetPointData().GetArray(0)

            for idx in range(imageSize):
                intensity.SetValue((layer * imageSize) + idx, rgbaArray.GetValue(idx*4))
                opacity.SetValue((layer * imageSize) + idx, rgbaArray.GetValue(idx*4 + 3))

            with open(depthPaths[layerNames[layer]], 'rb') as depthFile:
                depthArrays.append(depthFile.read())

        # Apply pixel sorting
        destOrder = vtkUnsignedCharArray()
        destOrder.SetNumberOfComponents(1)
        destOrder.SetNumberOfTuples(numberOfValues)

        opacityOrder = vtkUnsignedCharArray()
        opacityOrder.SetNumberOfComponents(1)
        opacityOrder.SetNumberOfTuples(numberOfValues)

        intensityOrder = vtkUnsignedCharArray()
        intensityOrder.SetNumberOfComponents(1)
        intensityOrder.SetNumberOfTuples(numberOfValues)

        for pixelIdx in range(imageSize):
            depthStack = []
            for depthArray in depthArrays:
                depthStack.append((depthArray[pixelIdx], len(depthStack)))
            depthStack.sort(key=lambda tup: tup[0])

            for destLayerIdx in range(len(depthStack)):
                sourceLayerIdx = depthStack[destLayerIdx][1]

                # Copy Idx
                destOrder.SetValue((imageSize * destLayerIdx) + pixelIdx, sourceLayerIdx)
                opacityOrder.SetValue((imageSize * destLayerIdx) + pixelIdx, opacity.GetValue((imageSize * sourceLayerIdx) + pixelIdx))
                intensityOrder.SetValue((imageSize * destLayerIdx) + pixelIdx, intensity.GetValue((imageSize * sourceLayerIdx) + pixelIdx))

        with open(os.path.join(directory, 'alpha.uint8'), 'wb') as f:
            f.write(buffer(opacityOrder))

        with open(os.path.join(directory, 'intensity.uint8'), 'wb') as f:
            f.write(buffer(intensityOrder))

        with open(os.path.join(directory, 'order.uint8'), 'wb') as f:
            f.write(buffer(destOrder))


class SortedCompositeDataSetBuilder(VolumeCompositeDataSetBuilder):
    def __init__(self, location, cameraInfo, metadata={}, sections={}):
        VolumeCompositeDataSetBuilder.__init__(self, location, 'image/png', cameraInfo, metadata, sections)
        self.dataHandler.addTypes('sorted-composite', 'rgba')

        # Register order and color textures
        self.layerScalars = []
        self.dataHandler.registerData(name='order',     type='array', fileName='/order.uint8')
        self.dataHandler.registerData(name='alpha',     type='array', fileName='/alpha.uint8')
        self.dataHandler.registerData(name='intensity', type='array', fileName='/intensity.uint8', categories=['intensity'])

    def start(self, renderWindow, renderer):
        VolumeCompositeDataSetBuilder.start(self, renderWindow, renderer)
        imageSize = self.renderWindow.GetSize()
        self.dataConverter = ConvertVolumeStackToSortedStack(imageSize[0], imageSize[1])

    def activateLayer(self, colorBy, scalar):
        VolumeCompositeDataSetBuilder.activateLayer(self, 'root', '%s' % scalar, colorBy)
        self.layerScalars.append(scalar)

    def writeData(self, mapper):
        VolumeCompositeDataSetBuilder.writeData(self, mapper)

        # Fill data pattern
        self.dataHandler.getDataAbsoluteFilePath('order')
        self.dataHandler.getDataAbsoluteFilePath('alpha')
        self.dataHandler.getDataAbsoluteFilePath('intensity')

    def stop(self, clean=True, compress=True):
        VolumeCompositeDataSetBuilder.stop(self, compress=False)

        # Go through all directories and convert them
        for root, dirs, files in os.walk(self.dataHandler.getBasePath()):
            for name in dirs:
                print('Process', os.path.join(root, name))
                self.dataConverter.convert(os.path.join(root, name))

        # Rename index.json to info_origin.json
        os.rename(os.path.join(self.dataHandler.getBasePath(), "index.json"), os.path.join(self.dataHandler.getBasePath(), "index_origin.json"))

        # Update index.json
        with open(os.path.join(self.dataHandler.getBasePath(), "index_origin.json"), "r") as infoFile:
            metadata = json.load(infoFile)
            metadata['SortedComposite'] = {
                'dimensions': metadata['CompositePipeline']['dimensions'],
                'layers': self.dataConverter.layers,
                'scalars': self.layerScalars[0:self.dataConverter.layers]
            }

            # Clean metadata
            dataToKeep = []
            del metadata['CompositePipeline']
            for item in metadata['data']:
                if item['name'] in ['order', 'alpha', 'intensity']:
                    dataToKeep.append(item)
            metadata['data'] = dataToKeep
            metadata['type'] = [ "tonic-query-data-model", "sorted-composite", "alpha" ]

            # Override index.json
            with open(os.path.join(self.dataHandler.getBasePath(), "index.json"), 'w') as newMetaFile:
                newMetaFile.write(json.dumps(metadata))

        # Clean temporary data
        if clean:
            for root, dirs, files in os.walk(self.dataHandler.getBasePath()):
                print('Clean', root)
                for name in files:
                    if '_rgb.png' in name or '_depth.uint8' in name or name == "index_origin.json":
                        os.remove(os.path.join(root, name))

        if compress:
            for root, dirs, files in os.walk(self.dataHandler.getBasePath()):
                print('Compress', root)
                for name in files:
                    if '.uint8' in name and '.gz' not in name:
                        with open(os.path.join(root, name), 'rb') as f_in:
                            with gzip.open(os.path.join(root, name + '.gz'), 'wb') as f_out:
                                shutil.copyfileobj(f_in, f_out)
                        os.remove(os.path.join(root, name))
